#include <stdio.h>
#include <stdlib.h>

#include "ts_msgs.h"
#include "input.h"
#include "hnet.h"

/* a pointer to the array of sequential cycles */
cycle *cs;

/* a pointer to the array of neurons (i.e. to the Hopfield network) */
neuron *hn;

/* the lengths of the corresponding arrays */
long cs_size, hn_size;

/* the number of connections in the Hopfield network */
long connections;

/* procedure sepcycle find all sequential cycles, stores them in
 * the array cs, and stores the orientation of internal edges in es[].r
 */
void sepcycle ( void ) 
{ 
	long i, cc, pe, pee, pi;
	int b, d, f, j, k, pj, s;
	long *st_bot, *st_top;

	/* allocate memory for the stack */
	st_bot = ( long * ) malloc ( es_size * sizeof ( long ) );
	if ( !st_bot ) ts_error ( "sepcycle", 3 );
	
	/* allocate memory for sequential cycles */
	cs = ( cycle * ) malloc ( es_size * sizeof ( cycle ) );
	if ( !cs ) ts_error ( "sepcycle", 3 );
	
	/* initialize attributes of internal edges */
	cc = -1;
	for ( i = 0; i < es_size; i++ ) {
		es[i].f = (char) 0;
		es[i].c[0] = cc;
		es[i].r = (char) 0;
	}

	/* set the length of the array of cycles to zero */
	cs_size = 0;

	/* initialize the stack */
	st_top = st_bot;
	
	for ( pi = 0; pi < es_size; pi++ )

		/* test whether there is a direction, in which the edge pi has not been examined yet */
		if ( ( (int) es[pi].f ) != 3 ) {
			
			/* if pi has not been examined in both directions, push it on the stack */
			if ( st_top == st_bot + es_size ) ts_error ( "sepcycle", 4 );
			*st_top = pi;
			st_top++;

			/* repeat while the stack is nonempty */
			do {
				/* pop an edge from the stack for examination */
				st_top--;
				i = *st_top;
				
				/* repeat while there is a not-examined direction for i */
				while ( ( f = (int) es[i].f ) != 3 ) {

					/* get the direction */
					d = f % 2;

					/* set the direction as examined */
					es[i].f = (char) (f + d + 1);
					if ( (int) es[i].f > 3 )  ts_error ( "sepcycle", 7 );
					
					/* mark i as a part of triangle strip currently being examined */
					es[i].c[0] = ++cc;
					
					/* set a boolean to 1, i.e. the triangle strip does not intersect itself */
					b = 1;

					/* specify the first neighbour of i that we will try to append to the triangle strip */
					j = d;
					
					/* repeat until the whole cycle or the whole strip which i is part of has been gone
					 * through */
					do {
						/* if the neighbour pe is an internal edge,].  assign the index of i in its array
						 * of neighbours to pj */
						if ( ( pe = es[i].e[j] ) >= 0 ) pj = (int) ( es[i].ie[j] );
						
						/* if the triangle strip is a cycle, we should come back to i via its
						 * neighbour e[k] */
						j = j + 2;
						k = j % 4;

						/* set the turn */
						s = ( ( (int) es[i].r ) + d ) % 2;
						
						/* repeat while the neighbour is an internal edge or until we enclosed 
						 * the cycle */
						while ( pe >= 0 && ( pe != i || pj != k ) ) {

							/* switch the turn */
							s = 1 - s;

							/* if it is the first time pe is being examined (i.e. f=0, both
							 * directions are free), push pe on the stack */
							if ( ( f = (int) es[pe].f ) == 0 ) {
								if ( st_top == st_bot + es_size ) ts_error ( "sepcycle", 4 );
								*st_top = pe;
								st_top++;

								/* set the orientation of pe */
								es[pe].r = (char) ( (pj + s) % 2 );
							}
							
							/* for pe, set the appropriate direction as examined */
							es[pe].f = (char) ( f + pj % 2 + 1 );
							if ( (int) es[pe].f > 3 )  ts_error ( "sepcycle", 7 );
							
							/* reset the boolean if the strip intersects itself */
							if ( b ) b = ( es[pe].c[0] != cc );

							/* assign pe to the current triangle strip */
							es[pe].c[0] = cc;
							
							/* make a step along the strip */
							pj = ( pj + 2 ) % 4;
							if ( ( pee = es[pe].e[pj] ) >= 0 ) pj = (int) es[pe].ie[pj];
							pe = pee;
						}
					}
					while ( j < 4 && pe < 0 );
				
					/* if the current strip is a cycle that does not intersect itself, add it to cs */
					if ( b && pe >= 0 ) {
						(cs + cs_size) -> e = i;          /* for now, i is not the representative
										     edge of the cycle */
						(cs + cs_size) -> d = (char) d;
						cs_size++;
					}
				}
			} 
			while ( !( st_top == st_bot ) );
		}
	
	/* free the stack */
	free ( ( void * ) st_bot );

	/* reallocate the memory for cs so that cs does not occupy more space than it needs */
	if ( cs_size ) {
		cs = ( cycle * ) realloc ( ( void * ) cs, cs_size * sizeof ( cycle ) );
		if ( !cs ) ts_error ( "sepcycle", 3 );
	}
	else 
		free ( ( void * ) cs );
}

/* procedure local creates the neurons corresponding to internal edges 
 * and stores them to the array hn
 */
void local ( void ) 
{
	long i, j, pe, c_size;
	int k, m;

	/* allocate memory for the array of neurons */
	hn = ( neuron * ) malloc ( ( 2 * es_size + 2 * cs_size ) * sizeof ( neuron ) );
	if ( !hn ) ts_error ( "local", 3 );
	
	/* set the length of the array hn to zero */
	hn_size = 0;
	
	for ( i = 0; i < es_size; i++ )

		/* assign two neurons to each internal edge */
		for ( m = 0; m < 2; m++ ) {

			/* allocate memory for the array of neigbours and for the array of weights */
			hn[ hn_size ].c = ( long * ) malloc ( 11 * sizeof ( long ) );
			hn[ hn_size ].w = ( long * ) malloc ( 11 * sizeof ( long ) );
			if ( !hn[ hn_size ].c || !hn[ hn_size ].w ) ts_error ( "local", 3 );
			
			/* set the length of both the arrays to zero */
			c_size = 0;

			/* set the bias of the neuron */
			hn[ hn_size ].w0 = 5;
			
			/* the other neuron assigned to the same edge i is a neighbour */
			hn[ hn_size ].c[ c_size++ ] = 2 * i + 1 - m;
			
			/* find the remaining neighbours among those neurons that are assigned 
			 * to internal edges neighbouring the edge i */
			for ( k = 0; k < 4; k++ )
				if ( ( pe = es[i].e[k] ) >= 0 ) {
					hn[ hn_size ].c[ c_size++ ] = 2 * pe + m;
					if ( ( (int) (es[i].r) + k + m ) % 2 )
						hn[ hn_size ].c[ c_size++ ] = 2 * pe + 1 - m;
				}
			
			/* set the corresponding weights */
			for ( j = 0; j < c_size; j++ )
				hn[ hn_size ].w[j] = -7;
			
			/* set the number of neighbours attribute for the neuron */
			hn[ hn_size ].c_size = c_size;

			/* increase the number of neurons in hn */
			hn_size++;
		}
}
                           
/* procedure selredge assigns the representative edges to sequential cycles,
 * and for each internal edge, it stores the information of the cycles going
 * through it to es[].c[0], e[].c[1], and e[].f
 */
void selredge ( void )
{
	long cc, ee, i, pe, pee;
	int f, j, k;
	
	/* (re-)initialize attributes of internal edges */
	for ( i = 0; i < es_size; i++ ) {
		es[i].f = (char) 0;
		es[i].c[0] = es[i].c[1] = -1;
	}
	
	/* go through all sequential cycles and set the attributes of the internal edges you meet */
	for ( i = 0; i < cs_size; i++ ) {
		
		/* set this attribute of the cycle for later use */
		cs[i].b = (char) 0;
		
		/* initialize the current (pe) and terminating (ee) edge, and 
		 * specify the neigbouring edge for continuation */
		pe = ee = cs[i].e;
		j = (int) cs[i].d;
		
		/* repeat until you reach the terminating edge */
		do {
			/* get the direction */
			k = j % 2;

			/* the number of cycle that meets the edge pe in direction k is i */
			es[pe].c[k] = i;

			/* there is a cycle that meets pe in direction k */
			es[pe].f = (char) ( ( (int) es[pe].f ) + k + 1 );

			/* make a step along the cycle */
			pee = es[pe].e[j];
			j = ( ( (int) es[pe].ie[j] ) + 2 ) % 4;
			pe = pee;
		}
		while ( pe != ee );
	}
	
	/* select the representative edge for each sequential cycle */
	for ( cc = 0; cc < cs_size; cc++ ) {
		i = cc;
		
		do {
			/* break if the representative edge has already been selected for the cycle i */
			if ( (int) cs[i].b ) break;

			/* set the boolean to indicate that the representative edge has been selected for 
			 * the cycle i */
			cs[i].b = (char) 1;
			
			/* initialize the current (pe) and terminating (ee) edge, and
			 * specify the neighbouring edge for continuation */
			pe = ee = cs[i].e;
			j = (int) cs[i].d;
			
			/* skip all edges that are representative for some cycle and also those
			 * that are common edges of two cycles */
			while ( ( f = (int) es[pe].f ) > 2 ) {

				/* make a step along the cycle i */
				pee = es[pe].e[j];
				j = ( ( (int) es[pe].ie[j] ) + 2 ) % 4;
				
				/* break if all edges of the cycle i are either representative for some cycle
				 * or common for two cycles */
				if ( ( pe = pee ) == ee ) break;
			}
			
			/* if you came here by breaking the while loop above, you have to skip at least the
			 * edges that are representative for some cycle */
			while ( ( f = (int) es[pe].f ) == 4 ) {

				/* make a step along the cycle i */
				pee = es[pe].e[j];
				j = ( ( (int) es[pe].ie[j] ) + 2 ) % 4;
				pe = pee;
			}
			
			/* select pe as the representative edge for the cycle i */
			cs[i].e = pe;

			/* save the direction */
			cs[i].d = (char) ( j % 2 );

			/* mark the edge as representative */
			es[pe].f = (char) 4;
			
			/* if the selected representative edge is a part of another cycle, handle this
			 * cycle now */
			if ( f == 3 ) i = es[pe].c[ (j + 1) % 2 ];
		}
		/* repeat while the selected representative edge is a common edge of two cycles */
		while ( f > 2 );
	}
}

/* procedure global adds the remaining neurons to the array of neurons, and
 * thus finishes the creation of a Hopfield network */
void global ( void ) 
{
	long cn, ee, e1, e2, i, k, kn, pe, pee, rn;
	int j, m;
	long *a, a_size, c_size;

	/* for a cycle C, a is used temporarily as the array of internal 
	 * edges in B'_C, that is, the array of internal edges that are 
	 * boundary with respect to C with exclusion of two of them - the 
	 * two that are neigbours of the representative edge of C;
	 * allocate memory for this array */
	a = ( long * ) malloc ( es_size * sizeof ( long ) );
	if ( !a ) ts_error ( "global", 3 );
	
	/* take all the cycles successively */
	for ( i = 0, cn = hn_size; i < cs_size; i++, cn++ ) {
		
		/* reset the length of the array a */
		a_size = 0;
	
		/* initialize the current (pe) and terminating (ee) edge, 
		 * both will now be set to the representative edge of C;
		 * specify the neighbouring edge for continuation */
		pe = ee = cs[i].e;
		j = (int) cs[i].d;

		/* the neigbours of the representative edge, which do not belong to B'_C */
		e1 = es[pe].e[ 1-j ];
		e2 = es[pe].e[ 3-j ];

		/* repeat until you reach the terminating edge */
		do {
			/* get the appropriate boundary (with respect to C) edge */
			pee = es[pe].e[ ( 5 - j ) % 4 ];
			
			/* if it is internal and not a neighbour of the representative edge,
			 * add it to a */
			if ( pee >= 0 && pee != e1 && pee != e2 )
				a[ a_size++ ] = pee;
			
			/* make a step along the cycle */
			pee = es[pe].e[j];
			j = ( ( (int) es[pe].ie[j] ) + 2 ) % 4;
			pe = pee;
		}
		while ( pe != ee );
		
		/* sort the array numerically in ascending order */
		qsort ( a, a_size, sizeof ( long ), &compare_ord );
		
		/* allocate memory for the arrays of neighbours and weights of the neuron d_C */
		hn[ hn_size ].c = ( long * ) malloc ( ( 2 * a_size + 2 ) * sizeof ( long ) );
		hn[ hn_size ].w = ( long * ) malloc ( ( 2 * a_size + 2 ) * sizeof ( long ) );
		if ( !hn[ hn_size ].c || !hn[ hn_size ].w ) ts_error ( "global", 3 );
		
		/* d_C has no neighbour so far */
		c_size = 0;
		
		for ( k = 0; k < a_size; k++ )
			
			/* count each of distinct elements of a only once */
			if ( k == a_size - 1 || a[k] != a[k + 1] ) {

				/* the lesser index of the two corresponding to neurons associated with the edge a[k] */
				kn = 2 * a[k];

				/* handle both these two neurons */
				for ( m = 0; m < 2; m++ ) {

					/* change the bias appropriately */
					hn[ kn + m ].w0 -= 2;

					/* declare the newly created d_C to be a neighbour of hn[ kn + m ] 
					 * and assign a weight to their connection */
					if ( hn[ kn + m ].c_size == 11 ) ts_error ( "global", 5 );
					hn[ kn + m ].c[  hn[ kn + m ].c_size  ] = cn;
					hn[ kn + m ].w[  hn[ kn + m ].c_size++  ] = 2;

					/* declare hn[ kn + m ] to be a neighbour of d_C
					 * and assign the weight even at this side of the connection */
					if ( c_size == 2 * a_size + 2 ) ts_error ( "global", 5 );
					hn[ hn_size ].c[ c_size ] = kn + m;
					hn[ hn_size ].w[ c_size++ ] = 2;
				}
			}
		
		/* in a while, the neuron a_C will be created; declare it
		 * to be a neighbour of d_C and assign a weight to their connection */
		if ( c_size == 2 * a_size + 2 ) ts_error ( "global", 5 );
		hn[ hn_size ].c[ c_size ] = cn + 1;
		hn[ hn_size ].w[ c_size++ ] = 2;
		
		/* get the index of the neuron that is assigned to the representative edge of C and
		 * represents the right direction, i.e. of the neuron j_C */
		rn = 2 * ee + ( (int) cs[i].d + (int) es[ee].r ) % 2;
		
		/* declare j_C to be a neighbour of d_C and assign a weight to their connection */
		if ( c_size == 2 * a_size + 2 ) ts_error ( "global", 5 );
		hn[ hn_size ].c[ c_size ] = rn;
		hn[ hn_size ].w[ c_size++ ] = 7;

		/* set the number of neighbours attribute and the bias for d_C */
		hn[ hn_size ].c_size = c_size;
		hn[ hn_size ].w0 = -1;

		/* a new neuron, the d_C, has been added to hn */
		hn_size++;

		/* allocate memory for the arrays of neighbours and weights of a_C */
		hn[ hn_size ].c = ( long * ) malloc ( 2 * sizeof ( long ) );
		hn[ hn_size ].w = ( long * ) malloc ( 2 * sizeof ( long ) );
		if ( !hn[ hn_size ].c || !hn[ hn_size ].w ) ts_error ( "global", 3 );

		/* a_C will have two neighbours, d_C and j_C, connected via weights 2 and -2, respectively */
		hn[ hn_size ].c[0] = cn;
		hn[ hn_size ].w[0] = 2;
		hn[ hn_size ].c[1] = rn;
		hn[ hn_size ].w[1] = -2;

		/* set the number of neighbours attribute and the bias for a_C */
		hn[ hn_size ].c_size = 2;
		hn[ hn_size ].w0 = -1;

		/* a new neuron, the a_C, has been added to hn */
		hn_size++;
			
		/* change the bias of j_C appropriately */
		hn[ rn ].w0 -= 6;

		/* declare d_C and a_C to be neighbours of j_C and assign the right
		 * weights even at j_C's side of the connections */
		if ( hn[ rn ].c_size == 11 ) ts_error ( "global", 5 );
		hn[ rn ].c[  hn[ rn ].c_size  ] = cn;
		hn[ rn ].w[  hn[ rn ].c_size++  ] = 7;
		if ( hn[ rn ].c_size == 11 ) ts_error ( "global", 5 );
		hn[ rn ].c[  hn[ rn ].c_size  ] = ++cn;
		hn[ rn ].w[  hn[ rn ].c_size++  ] = -2;
	}

	/* free the array a */
	free ( ( void * ) a );
}

/* procedure printn prints the information about the newly created Hopfield network
 * into the file "hnet.txt"
 *
 *     fname     the name of the input .obj file
 */
void printn ( char *fname ) 
{
	long i, j;
	FILE *out;
	
	out = fopen ( "hnet.txt", "w" );
	if ( !out ) ts_error ( "printn", 2 );
	
	fprintf ( out, "Hopfield network with %u neurons, %u connections, and %u cycles for \"%s\"\n", hn_size, connections, cs_size, fname );
	
	for ( j = 0; j < hn_size; j++ ) {
		fprintf ( out, "\nNeuron no. %u with bias %d connected to neurons (via weights):\n", j + 1, hn[j].w0 );
		
		for ( i = 0; i < hn[j].c_size; i++ ) {
			fprintf ( out, "%u(%d)", hn[j].c[i] + 1, hn[j].w[i] );
			
			if ( i < hn[j].c_size - 1 ) {
				fprintf ( out, ", " );
				if ( i % 10 == 9 ) fprintf ( out, "\n" );
			}
		}
		
		fprintf ( out, "\n" );
	}
	
	if ( fclose ( out ) == EOF ) ts_warning ( "printn", 4 );
}

/* procedure getconnections computes the number of connections in the Hopfield network 
 */
void getconnections ( void )
{
	long i;
	
	connections = 0;
	
	for ( i = 0; i < hn_size; i++ ) connections += hn[i].c_size;

	connections /= 2;
}

/* procedure hnet creates the array of sequential cycles, selects the
 * representative edge for each of the cycles, and creates the Hopfield
 * network (an array of neurons); it writes the information about the Hopfield
 * network into the file "hnet.txt"
 *
 *     fname     the name of the input .obj file
 */
void hnet ( char *fname )
{ 
	/* create the array of sequential cycles */
	sepcycle ();

	/* create the neurons that are associated with internal edges */
	local ();

	/* select the representative edges for sequential cycles */
	selredge ();

	/* for each cycle C, create the neurons d_C and a_C */
	global ();

	/* compute the number of connections in the Hopfield network */
	getconnections ();
	
	/* print the information about the Hopfield network into the file "hnet.txt" */
	printn ( fname );
}
